// Hist: Output a histogram of a file
// This program is in the public domain

#include <iostream>
#include <algorithm>
#include <map>
#include <getopt.h>

using namespace std;

#include "record.h"

#define NAME    "hist"
#define VERSION "0.1"
#define DEFAULT_SCALE 1

struct histopts {
   double scale;
   int field;
   int hist_char;
   bool show_n;
   bool sort;
   string field_sep;
};

void handle_options(int argc, char *argv[], histopts* opts);
void hist(Record r, histopts *opts);
bool desc(pair<string, int> a, pair<string, int> b);
bool asc(pair<string, int> a, pair<string, int> b);
void showversion();
void showhelp();

bool (*comp)(pair<string, int>, pair<string, int>);

int main(int argc, char *argv[])
{
   int exit_status = 0;
   int files = 0;
   struct histopts opts;
   ifstream fp;

   handle_options(argc, argv, &opts);
   while (optind < argc) {
      files++;
      if (*argv[optind] == '-' && argv[optind][1] == '\0') {   /* Just a '-'; read stdin */
         puts("<stdin>:");
         hist(Record(&cin, opts.field_sep), &opts);
      }
      else {
         try { fp.open(argv[optind], ios::in); }
         catch (...) {
            cerr << NAME << ": Failed to open " << argv[optind] << " for reading.\n";
            exit_status = 1;
            continue;
         }
         cout << argv[optind] << ":\n";
         hist(Record(&fp, opts.field_sep), &opts);
         fp.close();
      }
      optind++;
   }

   if (files == 0)
      hist(Record(&cin, opts.field_sep), &opts);

   return exit_status;
}

void showversion()
{
   cout << NAME << " version " << VERSION << "\n";
   cout << "Written in 2004 by Matthew Crumley\n";
   cout << "This program is in the public domain.\n";
   exit(0);
}

void showhelp()
{
   cout << "Usage: " << NAME << " [OPTIONS] [FILE]...\n";
   cout << "Print a histogram of a specified field\n";
   cout << "\n";
   cout << "  -s <scale>       Set the histogram's scale\n";
   cout << "  -f <n>           Show a histogram for field 'n' (default 0=entire line)\n";
   cout << "  -c <character>   Use character to draw the histogram (default: *)\n";
   cout << "  -d <string>      Set the field separator\n";
   cout << "  -n               Show the actual number\n";
   cout << "  -S               Sort by frequency\n";
   cout << "  -r               Sort by frequency, reversed (implies -S)\n";
   cout << "\n";
   cout << "  -v\tShow version information and exit\n";
   cout << "  -h\tShow this help screen and exit\n";
   cout << "\n";
   cout << "With no FILE, or when FILE is -, read standard input.\n";
   cout << "\n";
   exit(0);
}

void handle_options(int argc, char *argv[], histopts* opts)
{
   int arg;

   opterr = 0;

   opts->field = 0;
   opts->scale = DEFAULT_SCALE;
   opts->hist_char = '*';
   opts->show_n = false;
   opts->sort = false;
   opts->field_sep = " ";
   comp = asc;

   while ((arg = getopt(argc, argv, ":d:s:f:c:nSrvh")) != -1) {
      switch (arg) {
      case 'd':
         opts->field_sep = optarg;
         break;
      case 's':
         opts->scale = atof(optarg);
         if (opts->scale == 0) opts->scale = DEFAULT_SCALE;
         break;
      case 'f':
         opts->field = atoi(optarg);
         break;
      case 'c':
         opts->hist_char = *optarg;
         break;
      case 'v':
         showversion();
         break;
      case 'n':
         opts->show_n = true;
         break;
      case 'r':
         comp = desc;
         // FALLTHROUGH
         // This prevents a -S from switching it back later
      case 'S':
         opts->sort = true;
         break;
      case 'h':
      case '?':
      default:
         showhelp();
         break;
      }
   }
   return;
}

void line(char c, int length)
{
   do {  // Always print something
      cout << c;
   } while (--length > 0);
   cout << '\n';
}

bool asc(pair<string, int> a, pair<string, int> b)
{
   return a.second < b.second;
}

bool desc(pair<string, int> a, pair<string, int> b)
{
   return b.second < a.second;
}

void hist(Record r, histopts *opts)
{
   map<string, int> number;
   map<string, int>::iterator np;

   while(r.read())
      if (r[opts->field].length())
         number[r[opts->field]]++;

   if (opts->sort) {    // Order by frequency
      vector< pair<string, int> > sorted;
      vector< pair<string, int> >::iterator sp;

      for (np = number.begin(); np != number.end(); np++)
         sorted.push_back(*np);
      sort(sorted.begin(), sorted.end(), comp);

      for (sp = sorted.begin(); sp != sorted.end(); sp++) {
         cout << sp->first << ':';
         if (opts->show_n)
            cout << '(' << sp->second << ')';
         line(opts->hist_char, (int)(sp->second * opts->scale));
      }
   }
   else {      // Do arbitrary order (alphabetical)
      for (np = number.begin(); np != number.end(); np++) {
         cout << np->first << ':';
         if (opts->show_n)
            cout << '(' << np->second << ')';
         line(opts->hist_char, (int)(np->second * opts->scale));
      }
   }

   return;
}
